#-----------------------------------------------------------------------
# common makefile for avr-gcc
#-----------------------------------------------------------------------

#-----------------------------------------------------------------------
#
#
#  avr-gcc.mak assume that directories organised as
#
#  proj/obj -.
#  proj/lst -+- working and output directories
#  proj/exe -'
#  proj/src - sources base directory
#
# Externally defined variables:
#
#  MCU = Target processor name
#  TRG = Target name (project name)
#  LIB = Additional libraries
#  CFLAGS  = C  project-dependent flags
#  ASFLAGS = AS project-dependent flags
#  LDFLAGS = LD project-dependent flags
#
#  PROJ_SRCTOP = Base directory of source files subtree. Must point to
#       proj/src and can be relative from makefile location.
#       For example, when main project makefile located at
#       proj/src/proj_main/  you can set PROJ_SRCTOP=..
#       For "simple" project with all sources and makefile placed to
#       proj/src you can set PROJ_SRCTOP=.
#
#  MODULES = Space separated source directories list (from PROJ_SRCTOP point)
#       All .c and .S files in this directories will be compiled and linked.
#

# do not print any commands
.SILENT :

#-----------------------------------------------------------------------

# Set default toolchain prefix if it does not defind in environment or project makefile
TOOL ?= LANG=en avr-

# Set tool names
CC	:= $(TOOL)gcc
AS	:= $(TOOL)gcc -x assembler-with-cpp
BIN	:= $(TOOL)objcopy
OBJCOPY := $(TOOL)objcopy
OBJDUMP := $(TOOL)objdump
SIZE	:= $(TOOL)size

ELFSIZE := $(SIZE) -A

COG	:= cog.py

#-----------------------------------------------------------------------
#
OBJDIR := $(PROJ_SRCTOP)/../obj
LSTDIR := $(PROJ_SRCTOP)/../lst
EXEDIR := $(PROJ_SRCTOP)/../exe

ifneq "$(SUBTARGET)" ""
OBJDIR := $(OBJDIR)/$(SUBTARGET)
EXEDIR := $(EXEDIR)/$(SUBTARGET)
LSTDIR := $(LSTDIR)/$(SUBTARGET)
endif

SRCDIRS := $(addprefix $(PROJ_SRCTOP)/,$(MODULES))
INCDIRS := $(SRCDIRS)

CSRC	:= $(EXTRA_CSRC) $(wildcard $(addsuffix /*.c,$(SRCDIRS)))
CXXSRC	:= $(EXTRA_CXXSRC) $(wildcard $(addsuffix /*.cpp,$(SRCDIRS)))
ASRC    := $(EXTRA_ASMCSRC) $(wildcard $(addsuffix /*.S,$(SRCDIRS)))
BAK      = $(wildcard $(addsuffix /*.*~*,$(SRCDIRS)))
OBJ	:= $(addprefix $(OBJDIR)/, $(notdir $(CSRC:.c=.o) $(CXXSRC:.cpp=.o) $(ASRC:.S=.o) $(EXTRA_OBJ) ))
LST	:= $(addprefix $(LSTDIR)/,$(notdir $(CSRC:.c=.lst) $(CPSRC:.cpp=.o) $(ASRC:.S=.lst) ))


ifndef TRGPGMEXT
TRGPGMEXT = hex eep
endif

TRGELF := $(EXEDIR)/$(TRG).elf
TRGPGM := $(addprefix $(EXEDIR)/$(TRG).,$(TRGPGMEXT))
TRGCOF := $(EXEDIR)/$(TRG).cof


#GENDEPFLAGS = -Wp,-MM,-MP,-MT,$(OBJDIR)/$(*F).o,-MF,$(OBJDIR)/$(@F).d
GENDEPFLAGS = -Wp,-MM,-MP,-MT,$@,-MF,$@.d

CDEFS	+= -DF_CPU=$(F_CPU)UL
ADEFS	+= -DF_CPU=$(F_CPU)

CFLAGS	+= $(CSTD) $(OPTIMIZE) $(CWARN) $(CDEFS) -mmcu=$(MCU)
CXXFLAGS += $(CXXSTD) $(OPTIMIZE) $(CWARN) $(CDEFS) -mmcu=$(MCU)
ASFLAGS	+= $(ADEFS) -mmcu=$(MCU)

ifneq ($(DBG),)
CFLAGS	+= $(DBG)
CXXFLAGS += $(DBG)
ASFLAGS	+= -Wa,$(DBG)
endif

ifneq ($(LINK_SCRIPT),)
LDFLAGS += -Wl,-T,$(LINK_SCRIPT)
#$(TRGELF) : $(LINK_SCRIPT)
endif
LDFLAGS += -mmcu=$(MCU) -Wl,-Map=$(LSTDIR)/$(TRG).map,--cref

ifeq ($(USE_GC),Y)
	CFLAGS += -ffunction-sections -fdata-sections
	CXXFLAGS += -ffunction-sections -fdata-sections
	LDFLAGS += -Wl,--gc-sections
endif

ifeq ($(USE_LTO),YES)
	CFLAGS		+= -flto
	CXXFLAGS	+= -flto
	LD_FLAGS	+= -flto $(OPTIMIZE)
endif

# search path for common sources and headers
vpath
vpath %.c $(SRCDIRS)
vpath %.cpp $(SRCDIRS)
vpath %.h $(SRCDIRS)
vpath %.S $(SRCDIRS)
vpath %.cog $(SRCDIRS)

#-----------------------------------------------------------------------
# targets of makefile
#
.PHONY : all debug printok

# default target - binaries for AVR programming
all:	begin $(TRGPGM) printok

debug:	begin $(TRGCOF) printok

begin:
	echo ======== Beginning of project  "$(TRG)"  processing ; \
	$(CC) --version

printok:  $(TRGELF)
	echo ======== All OK,  project size: ; $(ELFSIZE) $<

$(OBJ):    | $(OBJDIR) $(LSTDIR)
$(TRGELF): | $(EXEDIR) $(LSTDIR)
dump:      | $(LSTDIR)

$(OBJDIR):
	mkdir -p $(OBJDIR)

$(EXEDIR):
	mkdir -p $(EXEDIR)

$(LSTDIR):
	mkdir -p $(LSTDIR)

#-----------------------------------------------------------------------
# compilation rules
#
%.c : %.lcd
	echo ====  Converting $< to source files; \
	$(PYLCD) $<

%.h : %.lcd
	echo ====  Converting $< to source files; \
	$(PYLCD) $<

%.cpp : %.cog
	echo ====  Generate $@ from $< ; $(COG) $< >$@ || rm $@

$(OBJDIR)/%.o : %.c
	echo ====  Compiling file $< ; \
	$(CC) -E $(CFLAGS)  $(addprefix -I,$(INCDIRS)) $(GENDEPFLAGS) $< ;\
	$(CC) -c $(CFLAGS)  $(addprefix -I,$(INCDIRS)) \
	-Wa,-ahlmsd=$(LSTDIR)/$(notdir $(<:.c=.lst)) $< -o $@

$(OBJDIR)/%.o : %.cpp
	echo ====  Compiling file $< ; \
	$(CC) -E $(CXXFLAGS)  $(addprefix -I,$(INCDIRS)) $(GENDEPFLAGS) $< ;\
	$(CC) -c $(CXXFLAGS)  $(addprefix -I,$(INCDIRS)) \
	-Wa,-ahlmsd=$(LSTDIR)/$(notdir $(<:.cpp=.lst)) $< -o $@

$(OBJDIR)/%.o : %.S
	echo ====  Compiling $< ; \
	$(AS) -E $(ASFLAGS) $(addprefix -I,$(INCDIRS)) $(GENDEPFLAGS) $< ; \
	$(AS) -c $(ASFLAGS) $(addprefix -I,$(INCDIRS)) \
		-Wa,-alm=$(LSTDIR)/$(notdir $(<:.S=.lst)) $< -o $@

%.s : %.c
	echo ====  Generate assembly source from $< ; \
	$(CC) -S -fverbose-asm $(CFLAGS) $(addprefix -I,$(INCDIRS)) $< -o $(<:.c=.s)

%.s : %.cpp
	echo ====  Generate assembly source from $< ; \
	$(CC) -S -fverbose-asm $(CXXFLAGS) $(addprefix -I,$(INCDIRS)) $< -o $(<:.cpp=.s)

# preprocess only
%.E : %.c
	echo ====  Generate PREPROCESSED file from $< ; \
	$(CC) -E $(CFLAGS) $(addprefix -I,$(INCDIRS)) $< -o $(<:.c=.E)

%.E : %.cpp
	echo ====  Generate PREPROCESSED file from $< ; \
	$(CC) -E $(CXXFLAGS) $(addprefix -I,$(INCDIRS)) $< -o $(<:.cpp=.E)

%.E : %.S
	echo ====  Generate PREPROCESSED file from $< ; \
	$(CC) -E $(ASFLAGS) $(addprefix -I,$(INCDIRS)) $< -o $(<:.S=.E)

#-----------------------------------------------------------------------
# elf post-processing rules
#

# hex-files for programmer
%.hex: %.elf
	echo ====  extract $(notdir $@) ; \
	$(BIN) -O ihex -R .eeprom $< $@

%.eep: %.elf
	echo ====  extract $(notdir $@) ; \
	$(BIN) -O ihex -j .eeprom --set-section-flags=.eeprom="alloc,load" \
	--change-section-lma .eeprom=0 $< $@

# files for simulator
%.obj: %.elf
	echo ====  extract $(notdir $@) ; \
	$(BIN) -O avrobj -R .eeprom $< $@

%.cof: %.elf
	echo ====  extract $(notdir $@) ; \
	$(OBJTOOL) loadelf $< mapfile $*.map writecof $@

# "dizassemble" elf file
dump: $(TRGELF)
	echo ====  disassemble $(notdir $<) ; \
	$(OBJDUMP) -D -S -C $< >$(LSTDIR)/$(TRG).dump

#-----------------------------------------------------------------------
# link: instructions to create elf output file from object files
#
.SECONDARY : $(TRG).elf
.PRECIOUS : $(OBJ)
$(TRGELF) : $(OBJ)
	echo ====  Link to $(notdir $@) ;\
	$(CC) $^ $(LIB) $(LDFLAGS) -o $@

#-----------------------------------------------------------------------
#  dependencies

# all obj files depend on makefile because of makefile contain
#  some compiler's switches (optimisation, debug options, etc..)
$(OBJ) : Makefile

# dependencies inclusion
# if make target list not contain this targets - include dependencies
ifeq (,$(findstring clean,$(MAKECMDGOALS)))
 ifeq (,$(findstring cleanall,$(MAKECMDGOALS)))
  ifeq (,$(findstring directories,$(MAKECMDGOALS)))
#-include $(shell mkdir -p $(OBJDIR) 2>/dev/null) $(wildcard $(OBJDIR)/*.d)
-include $(wildcard $(OBJDIR)/*.d)
  endif
 endif
endif

#-----------------------------------------------------------------------
#  create temporary and ouptut directories
#
.PHONY : directories
directories:
	echo ====  Create working directories ;\
	mkdir -p $(OBJDIR) $(LSTDIR) $(EXEDIR)

.PHONY : dummy
dummy:

#-----------------------------------------------------------------------
# clean project directories from working & backup ('~' suffix) files
#
.PHONY : clean cleanall
clean:
	echo ====  Erasing working directories and indent/editor backup files; \
	rm -f $(TRGCOF) $(EXEDIR)/$(TRG).obj $(EXEDIR)/$(TRG).sym ; \
	rm -f $(TRGELF) ; \
	rm -f $(LSTDIR)/* ; \
	rm -f $(OBJDIR)/* ; \
	rm -f $(CSRC:.c=.s) ; \
	rm -f $(BAK)

cleanall: clean
	- echo ====  Remove target binaries and working directories ; \
	rm -f $(TRGPGM) ; \
	test -e $(EXEDIR) && rmdir $(EXEDIR) ; \
	test -e $(LSTDIR) && rmdir $(LSTDIR) ; \
	test -e $(OBJDIR) && rmdir $(OBJDIR)


#####	end of avr.make

